//
// This file is part of Nolimips, a MIPS simulator with unlimited registers
// Copyright (C) 2003, 2004, 2006, 2012 Benoit Perrot <benoit@lrde.epita.fr>
//
// Nolimips is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// Nolimips is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Nolimips.  If not, see <http://www.gnu.org/licenses/>.
//
#include <iostream>

#include "config.h"
#include "common.hh"

#include "misc/contract.hh"
#include "misc/has.hh"

#include "task/task.hh"
#include "task/task_register.hh"

namespace task
{

  TaskRegister &
  TaskRegister::instance()
  {
    static TaskRegister     unique;
    return unique;
  }

  // --------------------------------------------------------------------------
  // Task Register Engine
  // --------------------------------------------------------------------------

  // Register a task.
  void
  TaskRegister::register_task (const Task& task)
  {
    std::string		long_opt = "--" + task.long_opt();
    std::string		short_opt = "-" + task.short_opt();

    assertion(! has_key(tasks_, long_opt));
    tasks_[long_opt] = &task;
    if (short_opt.size () > 1)
      {
	assertion(! has_key(tasks_, short_opt));
	tasks_[short_opt] = &task;
      }
    modules_[task.module()].push_back(&task);
  }
  
  // Enable a task
  void
  TaskRegister::enable_task (const Task& task)
  {
    enabled_tasks_.push_back(&task);
  }
  void
  TaskRegister::enable_task (std::string task_name)
  {
    assertion(task_name.size() > 0);
    if (task_name[0] != '-')
      {
	if (task_name.size() > 1)
	  task_name = "--" + task_name;
	else
	  task_name =  "-" + task_name;
      }
    assertion(has_key(tasks_, task_name));
    enabled_tasks_.push_back(tasks_[task_name]);    
  }


  // Resolve dependencies by a simple topological sort
  void
  TaskRegister::resolve_dependencies(const Task* task,
				     std::set<Task const*>& visited_tasks,
				     std::list<Task const*>& sorted_tasks)
  {
    if (has(visited_tasks, task))
      return;
    visited_tasks.insert(task);
    
    for (Task::deps_type::const_iterator it = task->dependencies ().begin ();
	 it != task->dependencies ().end (); ++it)
      {
	std::string name = "--" + (*it);
	assertion(has_key(tasks_, name));
	const Task* dependency = tasks_[name];
	resolve_dependencies(dependency, visited_tasks, sorted_tasks);
      }
    sorted_tasks.push_back (task);
  }
  void
  TaskRegister::resolve_dependencies()
  {
    std::list<Task const*>	sorted_tasks;
    std::set<Task const*>	visited_tasks;
    for (std::list<Task const*>::const_iterator
	   it = enabled_tasks_.begin (); it != enabled_tasks_.end (); ++it)
      resolve_dependencies ((*it), visited_tasks, sorted_tasks);
    enabled_tasks_ = sorted_tasks;
  }


  // Find a task given its name
  TaskRegister::const_task_iterator
  TaskRegister::find_task(const std::string &option)
  {
    TaskRegister::const_task_iterator it = tasks_.lower_bound(option);

    if (it == tasks_.end() || it->first.find(option) != 0)
      {
	std::cerr << program_name << ": unrecognized option `"
		  << option << "'" << std::endl;
	it = tasks_.end();
      }
    else
      {
	TaskRegister::const_task_iterator next(it);
	++next;
	if (next != tasks_.end() && next->first.find(option) == 0)
	  {
	    std::cerr << program_name << ": ambiguous option `"
		      << option << "'" << std::endl;
	    it = tasks_.end();
	  }
      }

    return it;
  }

  // Parse arguments
  char*
  TaskRegister::parse_args (int argc, char *argv[])
  {
    char *res = 0;

    for (int i = 1; i < argc; ++i)
      {
	std::string arg(argv[i]);
	
	if ((1 < arg.size ()) && (arg[0] == '-'))
	  {
	    std::string *value = 0;

	    // Identify option
	    TaskRegister::const_task_iterator it(tasks_.end());
	    if (arg[1] == '-') // Long option
	      {
		// Split on `=' for value
		unsigned eq = arg.find('=');
		if (eq < arg.size())
		  {
		    // +1: do not include `=' in value
		    value = new std::string(arg, eq + 1);
		    arg.resize(eq);
		  }

		// Search for long option
		it = find_task(arg);
	      }
	    else // Short option(s)
	      {
		for (unsigned l = 1; l < arg.size() && !value; ++l)
		  {
		    std::string short_option("-");
		    short_option += arg[l];

		    // Search for short option
		    it = find_task(short_option);
		    
		    // Let last task fall through the remaining of the
		    // routine
		    if (l + 1 < arg.size())
		      if (it != tasks_.end())
			{
			  if (it->second->needs_value())
			    // Split here for value
			    value = new std::string(arg, l + 1);
			  else
			    // Enable task
			    enable_task(*(it->second));
			}
		  }
	      }
	    
	    // Enable (long or last short) task and set value if needed
	    if (it != tasks_.end())
	      {
		const Task *task = it->second;
		
		if (task->needs_value())
		  {
		    if (!value)
		      {
			++i;
			if (i < argc)
			  value = new std::string(argv[i]);
		      }

		    if (!value)
		      std::cerr
			<< program_name
			<< ": option `" << task->long_opt()
			<< "' takes a value" << std::endl;
		    else
		      {
			task->set_value(*value);
			enable_task(*task);
		      }
		  }
		else
		  {
		    if (value)
		      std::cerr
			<< program_name
			<< ": option `" << task->long_opt()
			<< "' does not take a value" << std::endl;
		    else
		      enable_task(*task);
		  }
	      }
	    else if (value)
	      std::cerr 
		<< program_name
		<< ": `"<< *value << "': unexpected value" << std::endl;

	    delete value;
	  }
	else
	  res = argv[i];
      }

    return res;
  }


  // Execute tasks, checking dependencies.
  void
  TaskRegister::execute (void)
  {
    resolve_dependencies();
    for (std::list<Task const *>::const_iterator
	   it = enabled_tasks_.begin (); it != enabled_tasks_.end (); ++it)
      (*it)->execute ();
  }


  // --------------------------------------------------------------------------
  // Task Register Options
  // --------------------------------------------------------------------------

  // Display version
  void
  TaskRegister::version(std::ostream& ostr)
  {
    ostr
      << PACKAGE" ("PACKAGE_NAME") "PACKAGE_VERSION << std::endl
      << "Written by Benoit Perrot." << std::endl
      << std::endl
      << "Copyright (C) 2003, 2004, 2005, 2006, 2008, 2009, 2010, 2012 "
      << "Benoit Perrot." << std::endl
      << PACKAGE" comes with ABSOLUTELY NO WARRANTY." << std::endl
      << "This is free software, and you are welcome to redistribute and "
      << "modify it " << std::endl
      << "under certain conditions; see source for details."
      << std::endl << std::endl;
  }

  // Display usage
  void
  TaskRegister::usage(std::ostream& ostr)
  {
    ostr << "Usage: " << program_name << " [-";
    for (std::map<std::string, Task const*>::const_iterator
	   it = tasks_.begin(); it != tasks_.end(); ++it)
      if ((*it).first.size() == 2)
	ostr << (*it).second->short_opt();
    ostr << "] ";

    unsigned len = 80;
    const char tab[] = "        ";
    for (std::map<std::string, Task const*>::const_iterator
	   it = tasks_.begin(); it != tasks_.end(); ++it)
      if ((*it).first.size () > 2)
	{
	  if (len + (*it).first.size () + 3 >= 80)
	    {
	      ostr << std::endl << tab;
	      len = sizeof (tab);
	    }
	  ostr << "[" << (*it).first << "] ";
	  len += (*it).first.size () + 3;
	}
    ostr << std::endl;
  }

  // Display help
  void
  TaskRegister::help(std::ostream& ostr)
  {
    for (std::map<std::string, std::list<Task const*> >::const_iterator
	   it = modules_.begin(); it != modules_.end(); ++it)
      {
	if ((*it).first.size () > 0)
	  ostr << " " << (*it).first << std::endl;
	for (std::list<Task const*>::const_iterator
	       jt = (*it).second.begin(); jt != (*it).second.end(); ++jt)
	  {
	    ostr << "  ";
	    if ((*jt)->short_opt ().size() > 0)
	      ostr << "-" << (*jt)->short_opt() << ", ";
	    else
	      ostr << "    ";
	    ostr << "--" << (*jt)->long_opt();
 	    if ((*jt)->needs_value())
 	      ostr << "=ARG";
	    if ((*jt)->long_opt().size () <= 8)
	      ostr << '\t';
	    ostr << '\t' << (*jt)->description() << std::endl;
	  }
	ostr << std::endl;
      }
    ostr << "Reports bugs to "PACKAGE_BUGREPORT << std::endl;
  }

  // Display tasks selection
  void
  TaskRegister::selection(std::ostream& ostr)
  {
    ostr << "Sorted tasks selection:" << std::endl;
    for (std::list<Task const*>::const_iterator
	   it = enabled_tasks_.begin (); it != enabled_tasks_.end (); ++it)
      ostr << "\t* " << (*it)->long_opt() << std::endl;
  }

} // namespace task
